import json
import logging
import os
from datetime import date
from enum import StrEnum
from functools import lru_cache
from pathlib import Path
from typing import Annotated, Any, Callable, ClassVar, Optional, Self

import tiktoken
from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from tokenizers import Tokenizer
from typing_extensions import deprecated

from unique_toolkit._common.pydantic_helpers import get_configuration_dict
from unique_toolkit.language_model.schemas import LanguageModelTokenLimits

_LOGGER = logging.getLogger(__name__)


TypeEncoder = Callable[[str], list[int]]
TypeDecoder = Callable[[list[int]], str]


@lru_cache(maxsize=10)
def get_tokenizer_from_path(path: Path) -> Tokenizer:
    if not path.exists():
        raise FileNotFoundError(f"Tokenizer not found: {path}")
    return Tokenizer.from_file(str(path))


@lru_cache(maxsize=10)
def get_tokenizer_from_tiktoken(tokenizer_name: str) -> tiktoken.Encoding:
    return tiktoken.get_encoding(tokenizer_name)


class LanguageModelName(StrEnum):
    AZURE_GPT_35_TURBO_0125 = "AZURE_GPT_35_TURBO_0125"
    AZURE_GPT_4_0613 = "AZURE_GPT_4_0613"
    AZURE_GPT_4_32K_0613 = "AZURE_GPT_4_32K_0613"
    AZURE_GPT_4_TURBO_2024_0409 = "AZURE_GPT_4_TURBO_2024_0409"
    AZURE_GPT_5_2025_0807 = "AZURE_GPT_5_2025_0807"
    AZURE_GPT_5_MINI_2025_0807 = "AZURE_GPT_5_MINI_2025_0807"
    AZURE_GPT_5_NANO_2025_0807 = "AZURE_GPT_5_NANO_2025_0807"
    AZURE_GPT_5_CHAT_2025_0807 = "AZURE_GPT_5_CHAT_2025_0807"
    AZURE_GPT_5_PRO_2025_1006 = "AZURE_GPT_5_PRO_2025_1006"
    AZURE_GPT_51_2025_1113 = "AZURE_GPT_51_2025_1113"
    AZURE_GPT_51_THINKING_2025_1113 = "AZURE_GPT_51_THINKING_2025_1113"
    AZURE_GPT_51_CHAT_2025_1113 = "AZURE_GPT_51_CHAT_2025_1113"
    AZURE_GPT_51_CODEX_2025_1113 = "AZURE_GPT_51_CODEX_2025_1113"
    AZURE_GPT_51_CODEX_MINI_2025_1113 = "AZURE_GPT_51_CODEX_MINI_2025_1113"
    AZURE_GPT_52_2025_1211 = "AZURE_GPT_52_2025_1211"
    AZURE_GPT_52_CHAT_2025_1211 = "AZURE_GPT_52_CHAT_2025_1211"
    AZURE_GPT_4o_2024_0513 = "AZURE_GPT_4o_2024_0513"
    AZURE_GPT_4o_2024_0806 = "AZURE_GPT_4o_2024_0806"
    AZURE_GPT_4o_2024_1120 = "AZURE_GPT_4o_2024_1120"
    AZURE_GPT_4o_MINI_2024_0718 = "AZURE_GPT_4o_MINI_2024_0718"
    AZURE_o1_MINI_2024_0912 = "AZURE_o1_MINI_2024_0912"
    AZURE_o1_2024_1217 = "AZURE_o1_2024_1217"
    AZURE_o3_MINI_2025_0131 = "AZURE_o3_MINI_2025_0131"
    AZURE_GPT_45_PREVIEW_2025_0227 = "AZURE_GPT_45_PREVIEW_2025_0227"
    AZURE_GPT_41_2025_0414 = "AZURE_GPT_41_2025_0414"
    AZURE_GPT_41_MINI_2025_0414 = "AZURE_GPT_41_MINI_2025_0414"
    AZURE_GPT_41_NANO_2025_0414 = "AZURE_GPT_41_NANO_2025_0414"
    AZURE_o3_2025_0416 = "AZURE_o3_2025_0416"
    AZURE_o4_MINI_2025_0416 = "AZURE_o4_MINI_2025_0416"
    AZURE_MODEL_ROUTER_2025_1118 = "AZURE_MODEL_ROUTER_2025_1118"
    ANTHROPIC_CLAUDE_3_7_SONNET = "litellm:anthropic-claude-3-7-sonnet"
    ANTHROPIC_CLAUDE_3_7_SONNET_THINKING = (
        "litellm:anthropic-claude-3-7-sonnet-thinking"
    )
    ANTHROPIC_CLAUDE_HAIKU_4_5 = "litellm:anthropic-claude-haiku-4-5"
    ANTHROPIC_CLAUDE_SONNET_4 = "litellm:anthropic-claude-sonnet-4"
    ANTHROPIC_CLAUDE_SONNET_4_5 = "litellm:anthropic-claude-sonnet-4-5"
    ANTHROPIC_CLAUDE_SONNET_4_6 = "litellm:anthropic-claude-sonnet-4-6"
    ANTHROPIC_CLAUDE_OPUS_4 = "litellm:anthropic-claude-opus-4"
    ANTHROPIC_CLAUDE_OPUS_4_1 = "litellm:anthropic-claude-opus-4-1"
    ANTHROPIC_CLAUDE_OPUS_4_5 = "litellm:anthropic-claude-opus-4-5"
    ANTHROPIC_CLAUDE_OPUS_4_6 = "litellm:anthropic-claude-opus-4-6"
    GEMINI_2_0_FLASH = "litellm:gemini-2-0-flash"
    GEMINI_2_5_FLASH = "litellm:gemini-2-5-flash"
    GEMINI_2_5_FLASH_LITE = "litellm:gemini-2-5-flash-lite"
    GEMINI_2_5_FLASH_LITE_PREVIEW_0617 = "litellm:gemini-2-5-flash-lite-preview-06-17"
    GEMINI_2_5_FLASH_PREVIEW_0520 = "litellm:gemini-2-5-flash-preview-05-20"
    GEMINI_2_5_PRO = "litellm:gemini-2-5-pro"
    GEMINI_2_5_PRO_EXP_0325 = "litellm:gemini-2-5-pro-exp-03-25"
    GEMINI_2_5_PRO_PREVIEW_0605 = "litellm:gemini-2-5-pro-preview-06-05"
    GEMINI_3_1_PRO_PREVIEW = "litellm:gemini-3-1-pro-preview"
    GEMINI_3_FLASH_PREVIEW = "litellm:gemini-3-flash-preview"
    GEMINI_3_PRO_PREVIEW = "litellm:gemini-3-pro-preview"
    GROK_4_1_FAST_NON_REASONING = "litellm:grok-4-1-fast-non-reasoning"
    GROK_4_1_FAST_REASONING = "litellm:grok-4-1-fast-reasoning"
    LITELLM_OPENAI_GPT_5 = "litellm:openai-gpt-5"
    LITELLM_OPENAI_GPT_5_MINI = "litellm:openai-gpt-5-mini"
    LITELLM_OPENAI_GPT_5_NANO = "litellm:openai-gpt-5-nano"
    LITELLM_OPENAI_GPT_5_CHAT = "litellm:openai-gpt-5-chat"
    LITELLM_OPENAI_GPT_5_PRO = "litellm:openai-gpt-5-pro"
    LITELLM_OPENAI_GPT_51 = "litellm:openai-gpt-5-1"
    LITELLM_OPENAI_GPT_51_THINKING = "litellm:openai-gpt-5-1-thinking"
    LITELLM_OPENAI_GPT_52 = "litellm:openai-gpt-5-2"
    LITELLM_OPENAI_GPT_52_THINKING = "litellm:openai-gpt-5-2-thinking"
    LITELLM_OPENAI_O1 = "litellm:openai-o1"
    LITELLM_OPENAI_O3 = "litellm:openai-o3"
    LITELLM_OPENAI_O3_DEEP_RESEARCH = "litellm:openai-o3-deep-research"
    LITELLM_OPENAI_O3_PRO = "litellm:openai-o3-pro"
    LITELLM_OPENAI_O4_MINI = "litellm:openai-o4-mini"
    LITELLM_OPENAI_O4_MINI_DEEP_RESEARCH = "litellm:openai-o4-mini-deep-research"
    LITELLM_OPENAI_GPT_4_1_MINI = "litellm:openai-gpt-4-1-mini"
    LITELLM_OPENAI_GPT_4_1_NANO = "litellm:openai-gpt-4-1-nano"
    LITELLM_DEEPSEEK_R1 = "litellm:deepseek-r1"
    LITELLM_DEEPSEEK_V3 = "litellm:deepseek-v3-1"
    LITELLM_QWEN_3 = "litellm:qwen-3-235B-A22B"
    LITELLM_QWEN_3_THINKING = "litellm:qwen-3-235B-A22B-thinking"


class EncoderName(StrEnum):
    O200K_BASE = "o200k_base"
    CL100K_BASE = "cl100k_base"
    QWEN = "qwen"
    DEEPSEEK = "deepseek"

    @property
    def tokenizer_path(self) -> Path:
        base_path = Path(__file__).parent.parent / "_common" / "token" / "tokenizers"
        return base_path / self.value / "tokenizer.json"

    def get_encoder(self) -> TypeEncoder:
        if self.value in {"cl100k_base", "o200k_base"}:
            tokenizer = get_tokenizer_from_tiktoken(self.value)
            return tokenizer.encode

        tokenizer = get_tokenizer_from_path(self.tokenizer_path)
        return lambda text: tokenizer.encode(text).ids

    def get_decoder(self) -> TypeDecoder:
        if self.value in {"cl100k_base", "o200k_base"}:
            tokenizer = get_tokenizer_from_tiktoken(self.value)
            return tokenizer.decode

        tokenizer = get_tokenizer_from_path(self.tokenizer_path)
        return tokenizer.decode


def get_encoder_name(model_name: LanguageModelName) -> EncoderName:
    LMN = LanguageModelName
    match model_name:
        case LMN.AZURE_GPT_35_TURBO_0125:
            return EncoderName.CL100K_BASE
        case (
            LMN.AZURE_GPT_4_0613
            | LMN.AZURE_GPT_4_32K_0613
            | LMN.AZURE_GPT_4_TURBO_2024_0409
        ):
            return EncoderName.CL100K_BASE
        case (
            LMN.AZURE_GPT_4o_2024_0513
            | LMN.AZURE_GPT_4o_2024_0806
            | LMN.AZURE_GPT_4o_MINI_2024_0718
            | LMN.AZURE_GPT_4o_2024_1120
            | LMN.AZURE_GPT_5_2025_0807
            | LMN.AZURE_GPT_5_MINI_2025_0807
            | LMN.AZURE_GPT_5_NANO_2025_0807
            | LMN.AZURE_GPT_5_CHAT_2025_0807
            | LMN.AZURE_GPT_5_PRO_2025_1006
            | LMN.AZURE_GPT_51_2025_1113
            | LMN.AZURE_GPT_51_THINKING_2025_1113
            | LMN.AZURE_GPT_51_CHAT_2025_1113
            | LMN.AZURE_GPT_51_CODEX_2025_1113
            | LMN.AZURE_GPT_51_CODEX_MINI_2025_1113
            | LMN.AZURE_GPT_52_2025_1211
            | LMN.AZURE_GPT_52_CHAT_2025_1211
            | LMN.AZURE_MODEL_ROUTER_2025_1118
            | LMN.LITELLM_OPENAI_GPT_5
            | LMN.LITELLM_OPENAI_GPT_5_MINI
            | LMN.LITELLM_OPENAI_GPT_5_NANO
            | LMN.LITELLM_OPENAI_GPT_5_CHAT
            | LMN.LITELLM_OPENAI_GPT_5_PRO
            | LMN.LITELLM_OPENAI_GPT_51
            | LMN.LITELLM_OPENAI_GPT_51_THINKING
            | LMN.LITELLM_OPENAI_GPT_52
            | LMN.LITELLM_OPENAI_GPT_52_THINKING
            | LMN.LITELLM_OPENAI_O1
            | LMN.LITELLM_OPENAI_O3
            | LMN.LITELLM_OPENAI_O3_DEEP_RESEARCH
            | LMN.LITELLM_OPENAI_O4_MINI
            | LMN.LITELLM_OPENAI_O4_MINI_DEEP_RESEARCH
            | LMN.LITELLM_OPENAI_GPT_4_1_MINI
            | LMN.LITELLM_OPENAI_GPT_4_1_NANO
            | LMN.LITELLM_OPENAI_O3_PRO
        ):
            return EncoderName.O200K_BASE
        case _:
            print(
                f"{model_name} is not supported. Please add encoder information. Using default"
            )
            return EncoderName.CL100K_BASE


class LanguageModelProvider(StrEnum):
    AZURE = "AZURE"
    CUSTOM = "CUSTOM"
    LITELLM = "LITELLM"


class ModelCapabilities(StrEnum):
    CHAT_COMPLETIONS_API = "chat_completions_api"
    FUNCTION_CALLING = "function_calling"
    PARALLEL_FUNCTION_CALLING = "parallel_function_calling"
    REASONING = "reasoning"
    REPRODUCIBLE_OUTPUT = "reproducible_output"
    RESPONSES_API = "responses_api"
    STRUCTURED_OUTPUT = "structured_output"
    STREAMING = "streaming"
    VISION = "vision"


class TemperatureBounds(BaseModel):
    min_temperature: float
    max_temperature: float


def _load_custom_encoder(tokenizer_name: str) -> TypeEncoder:
    """Load a custom tokenizer from UNIQUE_CUSTOM_TOKENIZERS_PATH."""
    custom_path = os.getenv("UNIQUE_CUSTOM_TOKENIZERS_PATH")
    if not custom_path:
        raise ValueError(
            "UNIQUE_CUSTOM_TOKENIZERS_PATH must be set to use custom tokenizers"
        )

    tokenizer_path = Path(custom_path) / tokenizer_name / "tokenizer.json"

    tokenizer = get_tokenizer_from_path(tokenizer_path)
    return lambda text: tokenizer.encode(text).ids


def _load_custom_decoder(tokenizer_name: str) -> TypeDecoder:
    """Load a custom tokenizer from UNIQUE_CUSTOM_TOKENIZERS_PATH."""
    custom_path = os.getenv("UNIQUE_CUSTOM_TOKENIZERS_PATH")
    if not custom_path:
        raise ValueError(
            "UNIQUE_CUSTOM_TOKENIZERS_PATH must be set to use custom tokenizers"
        )
    tokenizer_path = Path(custom_path) / tokenizer_name / "tokenizer.json"
    tokenizer = get_tokenizer_from_path(tokenizer_path)
    return tokenizer.decode


class LanguageModelInfo(BaseModel):
    model_config = get_configuration_dict()
    name: (
        Annotated[str, Field(title="Custom Model Name")]
        | SkipJsonSchema[LanguageModelName]
    ) = Field(title="Model Name", default=LanguageModelName.AZURE_GPT_4o_2024_1120)
    provider: LanguageModelProvider = LanguageModelProvider.AZURE
    version: str = Field(title="Model Version", default="")

    encoder_name: EncoderName | Annotated[str, Field(title="Custom Encoder Name")] = (
        Field(default=EncoderName.CL100K_BASE, union_mode="left_to_right")
    )

    # TODO: Discuss if this is a sensible defaut
    token_limits: LanguageModelTokenLimits = LanguageModelTokenLimits(
        token_limit_input=7_000, token_limit_output=1_000
    )

    capabilities: list[ModelCapabilities] = [ModelCapabilities.STREAMING]

    info_cutoff_at: (
        Annotated[date, Field(title="Info Cutoff")]
        | Annotated[None, Field(title="Info Cutoff Unknown")]
    ) = None

    published_at: (
        Annotated[date, Field(title="Publishing Date")]
        | Annotated[None, Field(title="Publishing Date Unknown")]
    ) = None

    retirement_at: (
        Annotated[date, Field(title="Retirement Date")]
        | Annotated[None, Field(title="Retirement Date Unknown")]
    ) = date(2225, 12, 31)

    deprecated_at: (
        Annotated[date, Field(title="Deprecated Date")]
        | Annotated[None, Field(title="Deprecated Date Unknown")]
    ) = date(2225, 12, 31)

    retirement_text: str = "This model is no longer supported."

    temperature_bounds: (
        TemperatureBounds | Annotated[None, Field(title="Temperature Bounds Unknown")]
    ) = None

    default_options: dict[str, Any] = {}

    _ENV_VAR: ClassVar[str] = "LANGUAGE_MODEL_INFOS"

    def get_encoder(self) -> TypeEncoder:
        """Get an encode callable for this model's tokenizer."""
        if isinstance(self.encoder_name, EncoderName):
            return self.encoder_name.get_encoder()
        return _load_custom_encoder(self.encoder_name)

    def get_decoder(self) -> TypeDecoder:
        if isinstance(self.encoder_name, EncoderName):
            return self.encoder_name.get_decoder()
        return _load_custom_decoder(self.encoder_name)

    @classmethod
    @lru_cache(maxsize=1)
    def _load_from_env(cls) -> dict[str, dict]:
        """
        Load custom language model infos from environment variable.

        The environment variable should contain a JSON string with a dict of
        LanguageModelInfo-compatible dictionaries. The key is used for model lookup.

        Example:
            LANGUAGE_MODEL_INFOS='{"AZURE_GPT_4o_CUSTOM": {"name": "AZURE_GPT_4o_2024_1120",
            "provider": "AZURE", "version": "custom", "capabilities": ["function_calling",
            "streaming", "vision"], "token_limits": {"token_limit_input": 3000,
            "token_limit_output": 150}}}'

        Returns:
            A dictionary mapping model keys to their info dictionaries.
        """
        env_value = os.getenv(cls._ENV_VAR)
        if not env_value:
            return {}

        try:
            model_infos_dict = json.loads(env_value)
            if not isinstance(model_infos_dict, dict):
                _LOGGER.error(
                    f"{cls._ENV_VAR} must be a JSON dict of model info objects. "
                    f"Got {type(model_infos_dict).__name__} instead."
                )
                return {}

            # Validate each entry in the dictionary
            valid_model_infos: dict[str, dict] = {}
            for model_key, model_info in model_infos_dict.items():
                if not isinstance(model_info, dict):
                    _LOGGER.warning(
                        f"Skipping invalid model info entry '{model_key}' in {cls._ENV_VAR}: "
                        f"expected dict, got {type(model_info).__name__}"
                    )
                    continue

                valid_model_infos[model_key] = model_info

            _LOGGER.debug(
                f"Loaded {len(valid_model_infos)} custom language model infos from {cls._ENV_VAR}"
            )
            return valid_model_infos

        except json.JSONDecodeError:
            _LOGGER.error(
                f"Failed to parse {cls._ENV_VAR} as JSON. "
                "The environment variable should contain a valid JSON dict of model info objects.",
                exc_info=True,
            )
            return {}

    @classmethod
    def from_name(cls, model_name: LanguageModelName | str) -> Self:
        if isinstance(model_name, str) and model_name in [
            name.value for name in LanguageModelName
        ]:
            model_name = LanguageModelName(model_name)

        # Check environment variable first - env definitions take precedence
        env_model_infos = cls._load_from_env()
        model_name_str = (
            model_name.value
            if isinstance(model_name, LanguageModelName)
            else model_name
        )
        if model_name_str in env_model_infos.keys():
            try:
                return cls.model_validate(env_model_infos[model_name_str])
            except Exception:
                _LOGGER.warning(
                    f"Failed to parse model info for '{model_name_str}' from "
                    f"{cls._ENV_VAR}. Falling back to default definition.",
                    exc_info=True,
                )

        match model_name:
            case LanguageModelName.AZURE_GPT_35_TURBO_0125:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REPRODUCIBLE_OUTPUT,
                    ],
                    version="0125",
                    encoder_name=EncoderName.CL100K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=16385, token_limit_output=4096
                    ),
                    info_cutoff_at=date(2021, 9, 1),
                    published_at=date(2023, 1, 25),
                    retirement_at=date(5, 3, 31),
                )
            case LanguageModelName.AZURE_GPT_4_0613:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="0613",
                    encoder_name=EncoderName.CL100K_BASE,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                    ],
                    token_limits=LanguageModelTokenLimits(token_limit=8192),
                    info_cutoff_at=date(2021, 9, 1),
                    published_at=date(2023, 6, 13),
                    deprecated_at=date(2024, 10, 1),
                    retirement_at=date(2025, 6, 6),
                )
            case LanguageModelName.AZURE_GPT_4_32K_0613:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="1106-preview",
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                    ],
                    encoder_name=EncoderName.CL100K_BASE,
                    token_limits=LanguageModelTokenLimits(token_limit=32768),
                    info_cutoff_at=date(2021, 9, 1),
                    published_at=date(2023, 6, 13),
                    deprecated_at=date(2024, 10, 1),
                    retirement_at=date(2025, 6, 6),
                )
            case LanguageModelName.AZURE_GPT_5_2025_0807:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-08-07",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 10, 24),
                    published_at=date(2025, 8, 7),
                    deprecated_at=date(2026, 8, 7),
                    retirement_at=date(2026, 8, 7),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "minimal",
                    },
                )
            case LanguageModelName.AZURE_GPT_5_MINI_2025_0807:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-08-07",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 6, 24),
                    published_at=date(2025, 8, 7),
                    deprecated_at=date(2026, 8, 7),
                    retirement_at=date(2026, 8, 7),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "minimal",
                    },
                )
            case LanguageModelName.AZURE_GPT_5_NANO_2025_0807:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-08-07",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 5, 31),
                    published_at=date(2025, 8, 7),
                    deprecated_at=date(2026, 8, 7),
                    retirement_at=date(2026, 8, 7),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "minimal",
                    },
                )
            case LanguageModelName.AZURE_GPT_5_CHAT_2025_0807:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-08-07",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128000, token_limit_output=16384
                    ),
                    info_cutoff_at=date(2024, 10, 24),
                    published_at=date(2025, 8, 7),
                    deprecated_at=date(2026, 8, 7),
                    retirement_at=date(2026, 8, 7),
                )
            case LanguageModelName.AZURE_GPT_5_PRO_2025_1006:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-10-06",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 10, 30),
                    published_at=date(2025, 10, 6),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                )
            case LanguageModelName.AZURE_GPT_51_2025_1113:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-11-13",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 9, 30),
                    published_at=date(2025, 11, 13),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": None,
                    },
                )
            case LanguageModelName.AZURE_GPT_51_THINKING_2025_1113:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-11-13",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 9, 30),
                    published_at=date(2025, 11, 13),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "medium",
                    },
                )
            case LanguageModelName.AZURE_GPT_51_CHAT_2025_1113:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-11-13",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128_000, token_limit_output=16_384
                    ),
                    info_cutoff_at=date(2024, 9, 30),
                    published_at=date(2025, 11, 13),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "medium",
                    },
                )
            case LanguageModelName.AZURE_GPT_51_CODEX_2025_1113:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-11-13",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272_000, token_limit_output=128_000
                    ),
                    info_cutoff_at=date(2024, 9, 30),
                    published_at=date(2025, 11, 13),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "medium",
                    },
                )
            case LanguageModelName.AZURE_GPT_51_CODEX_MINI_2025_1113:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-11-13",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272_000, token_limit_output=128_000
                    ),
                    info_cutoff_at=date(2024, 9, 30),
                    published_at=date(2025, 11, 13),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "medium",
                    },
                )
            case LanguageModelName.AZURE_GPT_52_2025_1211:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-12-11",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=400_000, token_limit_output=128_000
                    ),
                    info_cutoff_at=date(2024, 9, 30),
                    published_at=date(2025, 12, 11),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": None,
                    },
                )
            case LanguageModelName.AZURE_GPT_52_CHAT_2025_1211:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.AZURE,
                    version="2025-12-11",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128_000, token_limit_output=16_384
                    ),
                    info_cutoff_at=date(2024, 9, 30),
                    published_at=date(2025, 12, 11),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "medium",
                    },
                )
            case LanguageModelName.AZURE_GPT_4_TURBO_2024_0409:
                return cls(
                    name=model_name,
                    encoder_name=EncoderName.CL100K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="turbo-2024-04-09",
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128000, token_limit_output=4096
                    ),
                    info_cutoff_at=date(2023, 12, 1),
                    published_at=date(2024, 4, 9),
                )
            case LanguageModelName.AZURE_GPT_4o_2024_0513:
                return cls(
                    name=model_name,
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2024-05-13",
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128_000, token_limit_output=4_096
                    ),
                    info_cutoff_at=date(2023, 10, 1),
                    published_at=date(2024, 5, 13),
                )
            case LanguageModelName.AZURE_GPT_4o_2024_0806:
                return cls(
                    name=model_name,
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2024-08-06",
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128_000, token_limit_output=16_384
                    ),
                    info_cutoff_at=date(2023, 10, 1),
                    published_at=date(2024, 8, 6),
                )
            case LanguageModelName.AZURE_GPT_4o_2024_1120:
                return cls(
                    name=model_name,
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2024-11-20",
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128_000, token_limit_output=16_384
                    ),
                    info_cutoff_at=date(2023, 10, 1),
                    published_at=date(2024, 11, 20),
                )
            case LanguageModelName.AZURE_GPT_4o_MINI_2024_0718:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2024-07-18",
                    encoder_name=EncoderName.O200K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128_000, token_limit_output=16_384
                    ),
                    info_cutoff_at=date(2023, 10, 1),
                    published_at=date(2024, 7, 18),
                )
            case LanguageModelName.AZURE_o1_MINI_2024_0912:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2024-09-12",
                    encoder_name=EncoderName.O200K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128_000, token_limit_output=65_536
                    ),
                    info_cutoff_at=date(2023, 10, 1),
                    published_at=date(2024, 9, 12),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                )
            case LanguageModelName.AZURE_o1_2024_1217:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2024-12-17",
                    encoder_name=EncoderName.O200K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=200_000, token_limit_output=100_000
                    ),
                    info_cutoff_at=date(2023, 10, 1),
                    published_at=date(2024, 12, 17),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                )
            case LanguageModelName.AZURE_o3_MINI_2025_0131:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2025-01-31",
                    encoder_name=EncoderName.O200K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=200_000, token_limit_output=100_000
                    ),
                    info_cutoff_at=date(2023, 10, 1),
                    published_at=date(2025, 1, 31),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                )
            case LanguageModelName.AZURE_o3_2025_0416:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2025-04-16",
                    encoder_name=EncoderName.O200K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=200_000, token_limit_output=100_000
                    ),
                    info_cutoff_at=date(2024, 5, 31),
                    published_at=date(2025, 4, 16),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                )
            case LanguageModelName.AZURE_o4_MINI_2025_0416:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2025-04-16",
                    encoder_name=EncoderName.O200K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=200_000, token_limit_output=100_000
                    ),
                    info_cutoff_at=date(2024, 5, 31),
                    published_at=date(2025, 4, 16),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                )
            case LanguageModelName.AZURE_GPT_45_PREVIEW_2025_0227:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2025-02-27",
                    encoder_name=EncoderName.O200K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128_000, token_limit_output=16_384
                    ),
                    info_cutoff_at=date(2023, 10, 1),
                    published_at=date(2025, 2, 27),
                )
            case LanguageModelName.AZURE_GPT_41_2025_0414:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2025-04-14",
                    encoder_name=EncoderName.O200K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_047_576, token_limit_output=32_768
                    ),
                    info_cutoff_at=date(2024, 5, 31),
                    published_at=date(2025, 4, 14),
                )
            case LanguageModelName.AZURE_GPT_41_MINI_2025_0414:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2025-04-14",
                    encoder_name=EncoderName.O200K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_047_576, token_limit_output=32_768
                    ),
                    info_cutoff_at=date(2024, 5, 31),
                    published_at=date(2025, 4, 14),
                )
            case LanguageModelName.AZURE_GPT_41_NANO_2025_0414:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2025-04-14",
                    encoder_name=EncoderName.O200K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_047_576, token_limit_output=32_768
                    ),
                    info_cutoff_at=date(2024, 5, 31),
                    published_at=date(2025, 4, 14),
                )
            case LanguageModelName.AZURE_MODEL_ROUTER_2025_1118:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.AZURE,
                    version="2025-11-18",
                    encoder_name=EncoderName.O200K_BASE,
                    token_limits=LanguageModelTokenLimits(
                        # Limit of smallest underlying model; router can route to larger contexts
                        token_limit_input=272_000,
                        token_limit_output=32_768,
                    ),
                )
            case LanguageModelName.ANTHROPIC_CLAUDE_3_7_SONNET:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="claude-3-7-sonnet",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        # Input limit is 200_000, we leave 20_000 tokens as buffer due to tokenizer mismatch
                        token_limit_input=180_000,
                        token_limit_output=64_000,
                    ),
                    info_cutoff_at=date(2024, 10, 31),
                    published_at=date(2025, 2, 24),
                )
            case LanguageModelName.ANTHROPIC_CLAUDE_3_7_SONNET_THINKING:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="claude-3-7-sonnet-thinking",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        # Input limit is 200_000, we leave 20_000 tokens as buffer due to tokenizer mismatch
                        token_limit_input=180_000,
                        token_limit_output=64_000,
                    ),
                    info_cutoff_at=date(2024, 10, 31),
                    published_at=date(2025, 2, 24),
                )
            case LanguageModelName.ANTHROPIC_CLAUDE_HAIKU_4_5:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="claude-haiku-4-5",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        # Input limit is 200_000, we leave 20_000 tokens as buffer due to tokenizer mismatch
                        token_limit_input=180_000,
                        token_limit_output=64_000,
                    ),
                    info_cutoff_at=date(2025, 2, 1),
                    published_at=date(2025, 10, 1),
                )
            case LanguageModelName.ANTHROPIC_CLAUDE_SONNET_4:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="claude-sonnet-4",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        # Input limit is 200_000, we leave 20_000 tokens as buffer due to tokenizer mismatch
                        token_limit_input=180_000,
                        token_limit_output=64_000,
                    ),
                    info_cutoff_at=date(2025, 3, 1),
                    published_at=date(2025, 5, 1),
                )
            case LanguageModelName.ANTHROPIC_CLAUDE_SONNET_4_5:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="claude-sonnet-4-5",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        # Input limit is 200_000, we leave 20_000 tokens as buffer due to tokenizer mismatch
                        token_limit_input=180_000,
                        token_limit_output=64_000,
                    ),
                    info_cutoff_at=date(2025, 7, 1),
                    published_at=date(2025, 9, 29),
                )
            case LanguageModelName.ANTHROPIC_CLAUDE_SONNET_4_6:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="claude-sonnet-4-6",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        # Input limit is 200_000, we leave 20_000 tokens as buffer due to tokenizer mismatch
                        token_limit_input=180_000,
                        token_limit_output=64_000,
                    ),
                    info_cutoff_at=date(2026, 1, 1),
                    published_at=date(2026, 2, 17),
                )
            case LanguageModelName.ANTHROPIC_CLAUDE_OPUS_4:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="claude-opus-4",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        # Input limit is 200_000, we leave 20_000 tokens as buffer due to tokenizer mismatch
                        token_limit_input=180_000,
                        token_limit_output=32_000,
                    ),
                    info_cutoff_at=date(2025, 3, 1),
                    published_at=date(2025, 5, 1),
                )
            case LanguageModelName.ANTHROPIC_CLAUDE_OPUS_4_1:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="claude-opus-4-1",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        # Input limit is 200_000, we leave 20_000 tokens as buffer due to tokenizer mismatch
                        token_limit_input=180_000,
                        token_limit_output=32_000,
                    ),
                    info_cutoff_at=date(2025, 3, 1),
                    published_at=date(2025, 5, 1),
                )
            case LanguageModelName.ANTHROPIC_CLAUDE_OPUS_4_5:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="claude-opus-4-5",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        # Input limit is 200_000, we leave 20_000 tokens as buffer due to tokenizer mismatch
                        token_limit_input=180_000,
                        token_limit_output=64_000,
                    ),
                    info_cutoff_at=date(2025, 8, 1),
                    published_at=date(2025, 11, 13),
                )
            case LanguageModelName.ANTHROPIC_CLAUDE_OPUS_4_6:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="claude-opus-4-6",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        # Input limit is 200_000, we leave 20_000 tokens as buffer due to tokenizer mismatch
                        token_limit_input=180_000,
                        token_limit_output=128_000,
                    ),
                    info_cutoff_at=date(2025, 8, 1),
                    published_at=date(2026, 2, 5),
                )
            case LanguageModelName.GEMINI_2_0_FLASH:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="gemini-2-0-flash",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_048_576, token_limit_output=8_192
                    ),
                    info_cutoff_at=date(2024, 8, 1),
                    published_at=date(2025, 2, 1),
                )
            case LanguageModelName.GEMINI_2_5_FLASH:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="gemini-2-5-flash",
                    encoder_name=EncoderName.O200K_BASE,  # TODO:Replace with LLM tokenizer
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_048_576, token_limit_output=65_536
                    ),
                    info_cutoff_at=date(2025, 1, day=1),
                    published_at=date(2025, 4, 1),
                )
            case LanguageModelName.GEMINI_2_5_FLASH_LITE:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="gemini-2-5-flash-lite",
                    encoder_name=EncoderName.O200K_BASE,  # TODO:Replace with LLM tokenizer
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_048_576, token_limit_output=65_536
                    ),
                    info_cutoff_at=date(2025, 1, day=1),
                    published_at=date(2025, 7, 1),
                )
            case LanguageModelName.GEMINI_2_5_FLASH_LITE_PREVIEW_0617:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="gemini-2-5-flash-lite-preview-06-17",
                    encoder_name=EncoderName.O200K_BASE,  # TODO:Replace with LLM tokenizer
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_000_000, token_limit_output=64_000
                    ),
                    info_cutoff_at=date(2025, 1, day=1),
                    published_at=date(2025, 6, 17),
                )
            case LanguageModelName.GEMINI_2_5_FLASH_PREVIEW_0520:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="gemini-2-5-flash-preview-05-20",
                    encoder_name=EncoderName.O200K_BASE,  # TODO:Replace with LLM tokenizer
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_048_576, token_limit_output=65_536
                    ),
                    info_cutoff_at=date(2025, 1, day=1),
                    published_at=date(2025, 4, 1),
                )
            case LanguageModelName.GEMINI_2_5_PRO:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="gemini-2-5-pro",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_048_576, token_limit_output=65_536
                    ),
                    info_cutoff_at=date(2025, 1, day=1),
                    published_at=date(2025, 6, 17),
                )
            case LanguageModelName.GEMINI_2_5_PRO_EXP_0325:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="gemini-2-5-pro-exp-0325",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_048_576, token_limit_output=65_536
                    ),
                    info_cutoff_at=date(2025, 1, day=1),
                    published_at=date(2025, 3, 1),
                )
            case LanguageModelName.GEMINI_2_5_PRO_PREVIEW_0605:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="gemini-2-5-pro-preview-06-05",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_048_576, token_limit_output=65_536
                    ),
                    info_cutoff_at=date(2025, 1, day=1),
                    published_at=date(2025, 6, 5),
                )
            case LanguageModelName.GEMINI_3_1_PRO_PREVIEW:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="gemini-3-1-pro-preview",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_048_576, token_limit_output=65_536
                    ),
                    info_cutoff_at=date(2025, 1, day=1),
                    published_at=date(2026, 1, 19),
                )
            case LanguageModelName.GEMINI_3_FLASH_PREVIEW:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="gemini-3-flash-preview",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_048_576, token_limit_output=65_536
                    ),
                    info_cutoff_at=date(2025, 1, day=1),
                    published_at=date(2025, 12, 17),
                )
            case LanguageModelName.GEMINI_3_PRO_PREVIEW:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="gemini-3-pro-preview",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with litellm
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_048_576, token_limit_output=65_536
                    ),
                    info_cutoff_at=date(2025, 1, day=1),
                    published_at=date(2025, 11, 13),
                )
            case LanguageModelName.GROK_4_1_FAST_NON_REASONING:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="grok-4-1-fast-non-reasoning",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with grok tokenizer
                    token_limits=LanguageModelTokenLimits(
                        # Context window is 2_000_000, we leave 200_000 tokens as buffer due to tokenizer mismatch
                        # Assign 90% for input and 10% for output
                        token_limit_input=int(1_800_000 * 0.9),
                        token_limit_output=int(1_800_000 * 0.1),
                    ),
                    info_cutoff_at=date(2024, 11, day=4),
                    published_at=date(2025, 11, 19),
                )
            case LanguageModelName.GROK_4_1_FAST_REASONING:
                return cls(
                    name=model_name,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    provider=LanguageModelProvider.LITELLM,
                    version="grok-4-1-fast-reasoning",
                    encoder_name=EncoderName.O200K_BASE,  # TODO: Update encoder with grok tokenizer
                    token_limits=LanguageModelTokenLimits(
                        # Context window is 2_000_000, we leave 200_000 tokens as buffer due to tokenizer mismatch
                        # Assign 90% for input and 10% for output
                        token_limit_input=int(1_800_000 * 0.9),
                        token_limit_output=int(1_800_000 * 0.1),
                    ),
                    info_cutoff_at=date(2024, 11, day=4),
                    published_at=date(2025, 11, 19),
                )
            case LanguageModelName.LITELLM_OPENAI_GPT_5:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="gpt-5",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 10, 24),
                    published_at=date(2025, 8, 7),
                    deprecated_at=date(2026, 8, 7),
                    retirement_at=date(2026, 8, 7),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "minimal",
                    },
                )
            case LanguageModelName.LITELLM_OPENAI_GPT_5_MINI:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="gpt-5-mini",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 6, 24),
                    published_at=date(2025, 8, 7),
                    deprecated_at=date(2026, 8, 7),
                    retirement_at=date(2026, 8, 7),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "minimal",
                    },
                )
            case LanguageModelName.LITELLM_OPENAI_GPT_5_NANO:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="gpt-5-nano",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 5, 31),
                    published_at=date(2025, 8, 7),
                    deprecated_at=date(2026, 8, 7),
                    retirement_at=date(2026, 8, 7),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "minimal",
                    },
                )
            case LanguageModelName.LITELLM_OPENAI_GPT_5_CHAT:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="gpt-5-chat",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128000, token_limit_output=16384
                    ),
                    info_cutoff_at=date(2024, 10, 24),
                    published_at=date(2025, 8, 7),
                    deprecated_at=date(2026, 8, 7),
                    retirement_at=date(2026, 8, 7),
                )
            case LanguageModelName.LITELLM_OPENAI_GPT_5_PRO:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-10-06",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 10, 30),
                    published_at=date(2025, 10, 6),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                )
            case LanguageModelName.LITELLM_OPENAI_GPT_51:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-11-13",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 9, 30),
                    published_at=date(2025, 11, 13),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": None,
                    },
                )
            case LanguageModelName.LITELLM_OPENAI_GPT_51_THINKING:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-11-13",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=272000, token_limit_output=128000
                    ),
                    info_cutoff_at=date(2024, 9, 30),
                    published_at=date(2025, 11, 13),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "medium",
                    },
                )
            case LanguageModelName.LITELLM_OPENAI_GPT_52:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-12-11",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=400_000, token_limit_output=128_000
                    ),
                    info_cutoff_at=date(2024, 9, 30),
                    published_at=date(2025, 12, 11),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": None,
                    },
                )
            case LanguageModelName.LITELLM_OPENAI_GPT_52_THINKING:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-12-11",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.PARALLEL_FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=400_000, token_limit_output=128_000
                    ),
                    info_cutoff_at=date(2024, 9, 30),
                    published_at=date(2025, 12, 11),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    default_options={
                        "reasoning_effort": "medium",
                    },
                )
            case LanguageModelName.LITELLM_OPENAI_O1:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2024-12-17",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=200_000, token_limit_output=100_000
                    ),
                    info_cutoff_at=date(2023, 10, 1),
                    published_at=date(2024, 12, 17),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                )
            case LanguageModelName.LITELLM_OPENAI_O3:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-04-16",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=200_000, token_limit_output=100_000
                    ),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                    published_at=date(2025, 4, 16),
                    info_cutoff_at=date(2024, 6, 1),
                )
            case LanguageModelName.LITELLM_OPENAI_O3_DEEP_RESEARCH:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-06-26",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=200_000, token_limit_output=100_000
                    ),
                    published_at=date(2025, 4, 16),
                    info_cutoff_at=date(2024, 6, 1),
                )
            case LanguageModelName.LITELLM_OPENAI_O3_PRO:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-06-10",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=200_000, token_limit_output=100_000
                    ),
                    published_at=date(2025, 6, 10),
                    info_cutoff_at=date(2024, 6, 1),
                )
            case LanguageModelName.LITELLM_OPENAI_O4_MINI:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-04-16",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=200_000, token_limit_output=100_000
                    ),
                    published_at=date(2025, 4, 16),
                    info_cutoff_at=date(2024, 6, 1),
                    temperature_bounds=TemperatureBounds(
                        min_temperature=1.0, max_temperature=1.0
                    ),
                )
            case LanguageModelName.LITELLM_OPENAI_O4_MINI_DEEP_RESEARCH:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-06-26",
                    encoder_name=EncoderName.O200K_BASE,
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.REASONING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=200_000, token_limit_output=100_000
                    ),
                    published_at=date(2025, 4, 16),
                    info_cutoff_at=date(2024, 6, 1),
                )
            case LanguageModelName.LITELLM_OPENAI_GPT_4_1_MINI:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-04-14",
                    encoder_name=EncoderName.O200K_BASE,
                    published_at=date(2025, 4, 14),
                    info_cutoff_at=date(2024, 6, 1),
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_047_576, token_limit_output=32_768
                    ),
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                )
            case LanguageModelName.LITELLM_OPENAI_GPT_4_1_NANO:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="2025-04-14",
                    encoder_name=EncoderName.O200K_BASE,
                    published_at=date(2025, 4, 14),
                    info_cutoff_at=date(2024, 6, 1),
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=1_047_576, token_limit_output=32_768
                    ),
                    capabilities=[
                        ModelCapabilities.CHAT_COMPLETIONS_API,
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.RESPONSES_API,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.VISION,
                    ],
                )
            case LanguageModelName.LITELLM_DEEPSEEK_R1:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="deepseek-r1",
                    encoder_name=EncoderName.DEEPSEEK,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.REASONING,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=64_000, token_limit_output=4_000
                    ),
                    published_at=date(2025, 1, 25),
                )
            case LanguageModelName.LITELLM_DEEPSEEK_V3:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="deepseek-v3-1",
                    encoder_name=EncoderName.DEEPSEEK,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.REASONING,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=128_000, token_limit_output=4_000
                    ),
                    published_at=date(2025, 8, 1),
                )
            case LanguageModelName.LITELLM_QWEN_3:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="qwen-3",
                    encoder_name=EncoderName.QWEN,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.REASONING,
                    ],
                    published_at=date(2025, 4, 29),
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=256_000, token_limit_output=32_768
                    ),
                )
            case LanguageModelName.LITELLM_QWEN_3_THINKING:
                return cls(
                    name=model_name,
                    provider=LanguageModelProvider.LITELLM,
                    version="qwen-3-thinking",
                    encoder_name=EncoderName.QWEN,
                    capabilities=[
                        ModelCapabilities.FUNCTION_CALLING,
                        ModelCapabilities.STRUCTURED_OUTPUT,
                        ModelCapabilities.STREAMING,
                        ModelCapabilities.REASONING,
                    ],
                    token_limits=LanguageModelTokenLimits(
                        token_limit_input=256_000, token_limit_output=32_768
                    ),
                    published_at=date(2025, 7, 25),
                )

            case _:
                if isinstance(model_name, LanguageModelName):
                    raise ValueError(
                        f"{model_name} is not supported. Please add model information in toolkit."
                    )

                return cls(
                    name=model_name,
                    version="custom",
                    provider=LanguageModelProvider.CUSTOM,
                )

    @property
    def display_name(self) -> str:
        """
        Returns the name of the model as a string.
        """

        if isinstance(self.name, LanguageModelName):
            return self.name.value
        else:
            return self.name


@deprecated(
    """
Use `LanguageModelInfo` instead of `LanguageModel`.

`LanguageModel` will be deprecated on 31.12.2025
""",
)
class LanguageModel:
    _info: ClassVar[LanguageModelInfo]

    def __init__(self, model_name: LanguageModelName | str):
        self._model_info = self.get_model_info(model_name)

    @property
    def info(self) -> LanguageModelInfo:
        """Return all infos about the model.

        - name
        - version
        - provider
        - encoder_name
        - token_limits
        - info_cutoff_at
        - published_at
        - retirement_at
        - deprecated_at
        - retirement_text
        """
        return self._model_info

    @property
    def name(self) -> LanguageModelName | str:
        """
        Returns the LanguageModelName of the model or the name string when it is a custom / not defined model.
        """
        return self._model_info.name

    @property
    def display_name(self) -> str:
        """
        Returns the name of the model as a string.
        """
        if isinstance(self._model_info.name, LanguageModelName):
            return self._model_info.name.name
        else:
            return self._model_info.name

    @property
    def version(self) -> Optional[str]:
        """
        Returns the version of the model.
        """
        return self._model_info.version

    @property
    def encoder_name(self) -> EncoderName | str:
        """
        Returns the encoder_name used for the model.
        """
        return self._model_info.encoder_name

    @property
    def token_limit(self) -> Optional[int]:
        """
        Returns the maximum number of tokens for the model.
        """
        if self._model_info.token_limits:
            return self._model_info.token_limits.token_limit

    @property
    def token_limit_input(self) -> Optional[int]:
        """
        Returns the maximum number of input tokens for the model.
        """
        if self._model_info.token_limits:
            return self._model_info.token_limits.token_limit_input

    @property
    def token_limit_output(self) -> Optional[int]:
        """
        Returns the maximum number of output tokens for the model.
        """
        if self._model_info.token_limits:
            return self._model_info.token_limits.token_limit_output

    @property
    def info_cutoff_at(self) -> Optional[date]:
        """
        Returns the date the model was last updated.
        """
        return self._model_info.info_cutoff_at

    @property
    def published_at(self) -> Optional[date]:
        """
        Returns the date the model was published.
        """
        return self._model_info.published_at

    @property
    def retirement_at(self) -> Optional[date]:
        """
        Returns the date the model will be retired.
        """
        return self._model_info.retirement_at

    @property
    def deprecated_at(self) -> Optional[date]:
        """
        Returns the date the model was deprecated.
        """
        return self._model_info.deprecated_at

    @property
    def retirement_text(self) -> Optional[str]:
        """
        Returns the text that will be displayed when the model is retired.
        """
        return self._model_info.retirement_text

    @property
    def provider(self) -> LanguageModelProvider:
        """
        Returns the provider of the model.
        """
        return self._model_info.provider

    @classmethod
    def get_model_info(cls, model_name: LanguageModelName | str) -> LanguageModelInfo:
        return LanguageModelInfo.from_name(model_name)

    @classmethod
    def list_models(cls) -> list[LanguageModelInfo]:
        """
        Returns a list of the infos of all available models.
        """

        return [
            LanguageModelInfo.from_name(model_name=name) for name in LanguageModelName
        ]



import logging
from typing import Unpack

from unique_toolkit.agentic.loop_runner._responses_stream_handler_utils import (
    responses_stream_response,
)
from unique_toolkit.agentic.loop_runner.base import (
    _ResponsesLoopIterationRunnerKwargs,
)
from unique_toolkit.language_model.schemas import ResponsesLanguageModelStreamResponse

_LOGGER = logging.getLogger(__name__)


async def handle_responses_last_iteration(
    **kwargs: Unpack[_ResponsesLoopIterationRunnerKwargs],
) -> ResponsesLanguageModelStreamResponse:
    _LOGGER.info("Reached last iteration, removing tools and producing final response")

    return await responses_stream_response(
        loop_runner_kwargs=kwargs,
        tools=None,
    )


async def handle_responses_normal_iteration(
    **kwargs: Unpack[_ResponsesLoopIterationRunnerKwargs],
) -> ResponsesLanguageModelStreamResponse:
    _LOGGER.info("Running loop iteration %d", kwargs["iteration_index"])

    return await responses_stream_response(loop_runner_kwargs=kwargs)


async def handle_responses_forced_tools_iteration(
    **kwargs: Unpack[_ResponsesLoopIterationRunnerKwargs],
) -> ResponsesLanguageModelStreamResponse:
    assert "tool_choices" in kwargs

    tool_choices = kwargs["tool_choices"]
    assert len(tool_choices) > 0

    _LOGGER.info("Forcing tools calls: %s", tool_choices)

    responses: list[ResponsesLanguageModelStreamResponse] = []

    for opt in tool_choices:
        responses.append(
            await responses_stream_response(loop_runner_kwargs=kwargs, tool_choice=opt)
        )

    # Merge responses and refs:
    tool_calls = []
    references = []
    for r in responses:
        if r.tool_calls:
            tool_calls.extend(r.tool_calls)
        references.extend(r.message.references)

    response = responses[0]
    response.tool_calls = tool_calls if len(tool_calls) > 0 else None
    response.message.references = references

    return response



from unique_toolkit.agentic.tools.schemas import ToolCallResponse
from unique_toolkit.content.schemas import ContentChunk, ContentReference


class tool_chunks:
    def __init__(self, name: str, chunks: list[ContentChunk]) -> None:
        self.name = name
        self.chunks = chunks


class ReferenceManager:
    """
    Manages content chunks and references extracted from tool responses.

    This class is responsible for:
    - Extracting and storing referenceable content chunks from tool responses.
    - Managing a collection of content chunks and their associated references.
    - Providing methods to retrieve, replace, and manipulate chunks and references.
    - Supporting the retrieval of the latest references and their corresponding chunks.

    Key Features:
    - Chunk Extraction: Extracts content chunks from tool responses and organizes them for reference.
    - Reference Management: Tracks references to content chunks and allows for easy retrieval.
    - Latest Reference Access: Provides methods to fetch the most recent references and their associated chunks.
    - Flexible Chunk Replacement: Allows for replacing the current set of chunks with a new list.
    - Reference-to-Chunk Mapping: Matches references to their corresponding chunks based on source IDs.

    The ReferenceManager serves as a utility for managing and linking content chunks with references, enabling efficient content tracking and retrieval.
    """

    def __init__(self):
        self._tool_chunks: dict[str, tool_chunks] = {}
        self._chunks: list[ContentChunk] = []
        self._references: list[list[ContentReference]] = []

    def extract_referenceable_chunks(
        self, tool_responses: list[ToolCallResponse]
    ) -> None:
        for tool_response in tool_responses:
            if not tool_response.content_chunks:
                continue
            self._chunks.extend(tool_response.content_chunks or [])
            self._tool_chunks[tool_response.id] = tool_chunks(
                tool_response.name, tool_response.content_chunks
            )

    def get_chunks(self) -> list[ContentChunk]:
        return self._chunks

    def get_tool_chunks(self) -> dict[str, tool_chunks]:
        return self._tool_chunks

    def get_chunks_of_all_tools(self) -> list[list[ContentChunk]]:
        return [tool_chunks.chunks for tool_chunks in self._tool_chunks.values()]

    def get_chunks_of_tool(self, tool_call_id: str) -> list[ContentChunk]:
        return self._tool_chunks.get(tool_call_id, tool_chunks("", [])).chunks

    def replace_chunks_of_tool(
        self, tool_call_id: str, chunks: list[ContentChunk]
    ) -> None:
        if tool_call_id in self._tool_chunks:
            self._tool_chunks[tool_call_id].chunks = chunks

    def replace(self, chunks: list[ContentChunk]):
        self._chunks = chunks

    def add_references(
        self,
        references: list[ContentReference],
    ):
        self._references.append(references)

    def get_references(
        self,
    ) -> list[list[ContentReference]]:
        return self._references

    def get_latest_references(
        self,
    ) -> list[ContentReference]:
        if not self._references:
            return []
        return self._references[-1]

    def get_latest_referenced_chunks(self) -> list[ContentChunk]:
        if not self._references:
            return []
        return self._get_referenced_chunks_from_references(self._references[-1])

    def _get_referenced_chunks_from_references(
        self,
        references: list[ContentReference],
    ) -> list[ContentChunk]:
        """
        Get _referenced_chunks by matching sourceId from _references with merged id and chunk_id from _chunks.
        """
        referenced_chunks: list[ContentChunk] = []
        for ref in references:
            for chunk in self._chunks:
                if ref.source_id == f"{chunk.id}_{chunk.chunk_id}":
                    referenced_chunks.append(chunk)
        return referenced_chunks



import json
from logging import getLogger
from typing import TYPE_CHECKING, Any, Awaitable, Callable, TypeVar

from pydantic import ValidationError

if TYPE_CHECKING:
    from fastapi import BackgroundTasks, FastAPI, Request, status
    from fastapi.responses import JSONResponse
else:
    try:
        from fastapi import BackgroundTasks, FastAPI, Request, status
        from fastapi.responses import JSONResponse
    except ImportError:
        FastAPI = None  # type: ignore[assignment, misc]
        Request = None  # type: ignore[assignment, misc]
        status = None  # type: ignore[assignment, misc]
        JSONResponse = None  # type: ignore[assignment, misc]
        BackgroundTasks = None  # type: ignore[assignment, misc]

from unique_toolkit.app.schemas import BaseEvent, ChatEvent, EventName
from unique_toolkit.app.unique_settings import UniqueSettings

logger = getLogger(__name__)


def default_event_handler(event: Any) -> int:
    logger.info("Event received at event handler")
    if status is not None:
        return status.HTTP_200_OK
    else:
        # No fastapi installed
        return 200


T = TypeVar("T", bound=BaseEvent)

EventHandlerType = Callable[[T], Awaitable[int]] | Callable[[T], int]


def build_unique_custom_app(
    *,
    title: str = "Unique Chat App",
    webhook_path: str = "/webhook",
    settings: UniqueSettings,
    event_handler: EventHandlerType = default_event_handler,
    event_constructor: Callable[..., T] = ChatEvent,
    subscribed_event_names: list[str] | None = None,
) -> "FastAPI":
    """Factory class for creating FastAPI apps with Unique webhook handling."""
    if FastAPI is None:
        raise ImportError(
            "FastAPI is not installed. Install it with: poetry install --with fastapi"
        )

    app = FastAPI(title=title)

    if subscribed_event_names is None:
        subscribed_event_names = [EventName.EXTERNAL_MODULE_CHOSEN]

    @app.get(path="/")
    async def health_check() -> JSONResponse:
        """Health check endpoint."""
        return JSONResponse(content={"status": "healthy", "service": title})

    @app.post(path=webhook_path)
    async def webhook_handler(
        request: Request, background_tasks: BackgroundTasks
    ) -> JSONResponse:
        """
        Webhook endpoint for receiving events from Unique platform.

        This endpoint:
        1. Verifies the webhook signature
        2. Constructs an event from the payload
        3. Calls the configured event handler
        """
        # Get raw body and headers
        body = await request.body()
        headers = dict(request.headers)

        from unique_toolkit.app.webhook import is_webhook_signature_valid

        if not is_webhook_signature_valid(
            headers=headers,
            payload=body,
            endpoint_secret=settings.app.endpoint_secret.get_secret_value(),
        ):
            return JSONResponse(
                status_code=status.HTTP_401_UNAUTHORIZED,
                content={"error": "Invalid webhook signature"},
            )

        try:
            event_data = json.loads(body.decode(encoding="utf-8"))
        except json.JSONDecodeError as e:
            logger.error(f"Error parsing event: {e}", exc_info=True)
            return JSONResponse(
                status_code=status.HTTP_400_BAD_REQUEST,
                content={"error": f"Invalid event format: {e.msg}"},
            )

        if event_data["event"] not in subscribed_event_names:
            return JSONResponse(
                status_code=status.HTTP_400_BAD_REQUEST,
                content={"error": "Not subscribed event"},
            )

        try:
            event = event_constructor(**event_data)
            if (
                settings.chat_event_filter_options
                and settings.chat_event_filter_options.assistant_ids
            ):
                if event.filter_event(
                    filter_options=settings.chat_event_filter_options
                ):
                    return JSONResponse(
                        status_code=status.HTTP_200_OK,
                        content={"error": "Event filtered out"},
                    )
        except ValidationError as e:
            # pydantic errors https://docs.pydantic.dev/2.10/errors/errors/
            logger.error(f"Validation error with model: {e.json()}", exc_info=True)
            raise e
        except ValueError as e:
            logger.error(f"Error deserializing event: {e}", exc_info=True)
            return JSONResponse(
                status_code=status.HTTP_400_BAD_REQUEST,
                content={"error": "Invalid event"},
            )

        # Run the task in background so that we don't block for long running tasks
        background_tasks.add_task(event_handler, event)
        return JSONResponse(
            status_code=status.HTTP_200_OK, content={"message": "Event received"}
        )

    return app



import logging

from typing_extensions import deprecated
from unique_sdk import (
    AgenticTable,
    AgenticTableSheetState,
    AgreementStatus,
    CellRendererTypes,
    FilterTypes,
    RowVerificationStatus,
    SelectionMethod,
)
from unique_sdk import AgenticTableCell as SDKAgenticTableCell
from unique_sdk.api_resources._agentic_table import ActivityStatus

from .schemas import (
    ArtifactType,
    LogEntry,
    MagicTableAction,
    MagicTableCell,
    MagicTableSheet,
    RowMetadataEntry,
)


class LockedAgenticTableError(Exception):
    pass


class AgenticTableService:
    """
    Provides methods to interact with the Agentic Table.

    Attributes:
        #event (ChatEvent): The ChatEvent object.
        logger (Optional[logging.Logger]): The logger object. Defaults to None.
    """

    def __init__(
        self,
        user_id: str,
        company_id: str,
        table_id: str,
        event_id: str | None = None,
        logger: logging.Logger = logging.getLogger(__name__),
    ):
        self._event_id = event_id
        self._user_id = user_id
        self._company_id = company_id
        self.table_id = table_id
        self.logger = logger

    async def set_cell(
        self,
        row: int,
        column: int,
        text: str,
        log_entries: list[LogEntry] | None = None,
    ):
        """
        Sets the value of a cell in the Agentic Table.

        Args:
            row (int): The row index.
            column (int): The column index.
            text (str): The text to set.
            log_entries (Optional[list[LogEntry]]): The log entries to set.
        """
        if log_entries is None:
            log_entries_new = []
        else:
            log_entries_new = [
                log_entry.to_sdk_log_entry() for log_entry in log_entries
            ]
        try:
            await AgenticTable.set_cell(
                user_id=self._user_id,
                company_id=self._company_id,
                tableId=self.table_id,
                rowOrder=row,
                columnOrder=column,
                text=text,
                logEntries=log_entries_new,
            )
        except Exception as e:
            self.logger.error(f"Error setting cell {row}, {column}: {e}.")

    async def get_cell(
        self, row: int, column: int, include_row_metadata: bool = True
    ) -> MagicTableCell:
        """
        Gets the value of a cell in the Agentic Table.

        Args:
            row (int): The row index.
            column (int): The column index.
            include_row_metadata (bool): Whether to include the row metadata. Defaults to True.

        Returns:
            MagicTableCell: The MagicTableCell object.

        """
        cell_data = await AgenticTable.get_cell(
            user_id=self._user_id,
            company_id=self._company_id,
            tableId=self.table_id,
            rowOrder=row,
            columnOrder=column,
            includeRowMetadata=include_row_metadata,  # type: ignore[arg-type]
        )
        return MagicTableCell.model_validate(cell_data)

    async def set_multiple_cells(
        self, cells: list[MagicTableCell], batch_size: int = 4000
    ):
        """
        Sets the values of multiple cells in the Agentic Table.

        Args:
            cells (list[MagicTableCell]): The cells to set sorted by row and column.
            batch_size (int): Number of cells to set in a single request.
        """
        for i in range(0, len(cells), batch_size):
            batch = cells[i : i + batch_size]
            await AgenticTable.set_multiple_cells(
                user_id=self._user_id,
                company_id=self._company_id,
                tableId=self.table_id,
                cells=[
                    SDKAgenticTableCell(
                        rowOrder=cell.row_order,
                        columnOrder=cell.column_order,
                        text=cell.text,
                    )
                    for cell in batch
                ],
            )

    async def set_activity(
        self,
        text: str,
        activity: MagicTableAction,
        status: ActivityStatus = ActivityStatus.IN_PROGRESS,
    ):
        """
        Sets the activity of the Agentic Table.

        Args:
            activity (str): The activity to set.
        """
        await AgenticTable.set_activity(
            user_id=self._user_id,
            company_id=self._company_id,
            tableId=self.table_id,
            activity=activity.value,  # type: ignore[arg-type]
            status=status.value,  # type: ignore[arg-type]
            text=text,
        )

    async def register_agent(self) -> None:
        """
        Registers the agent for the Agentic Table by updating the sheet state to PROCESSING.

        Raises:
            LockedAgenticTableError: If the Agentic Table is busy.
        """
        state = await AgenticTable.get_sheet_state(
            user_id=self._user_id,
            company_id=self._company_id,
            tableId=self.table_id,
        )
        if state == AgenticTableSheetState.IDLE:
            await AgenticTable.update_sheet_state(
                user_id=self._user_id,
                company_id=self._company_id,
                tableId=self.table_id,
                state=AgenticTableSheetState.PROCESSING,
            )
            return
        # If the sheet is not idle, we cannot register the agent
        raise LockedAgenticTableError(
            f"Agentic Table is busy. Cannot register agent {self._event_id or self.table_id}."
        )

    async def deregister_agent(self):
        """
        Deregisters the agent for the Agentic Table by updating the sheet state to IDLE.

        Raises:
            LockedAgenticTableError: If the Agentic Table is busy.
        """
        await AgenticTable.update_sheet_state(
            user_id=self._user_id,
            company_id=self._company_id,
            tableId=self.table_id,
            state=AgenticTableSheetState.IDLE,
        )

    async def set_artifact(
        self,
        artifact_type: ArtifactType,
        content_id: str,
        mime_type: str,
        name: str,
    ):
        """Upload/set report files to the Agentic Table.

        Args:
            artifact_type (ArtifactType): The type of artifact to set.
            content_id (str): The content ID of the artifact.
            mime_type (str): The MIME type of the artifact.
            name (str): The name of the artifact.
        """
        await AgenticTable.set_artifact(
            user_id=self._user_id,
            company_id=self._company_id,
            tableId=self.table_id,
            artifactType=artifact_type.value,
            contentId=content_id,
            mimeType=mime_type,
            name=name,
        )

    @deprecated("Use set_column_style instead.")
    async def set_column_width(self, column: int, width: int):
        await self.set_column_style(column=column, width=width)

    async def set_column_style(
        self,
        column: int,
        width: int | None = None,
        cell_renderer: CellRendererTypes | None = None,
        filter: FilterTypes | None = None,
        editable: bool | None = None,
    ):
        """
        Sets the style of a column in the Agentic Table.

        Args:
            column (int): The column index.
            width (int | None, optional): The width of the column. Defaults to None.
            cell_renderer (CellRenderer | None, optional): The cell renderer of the column. Defaults to None.
            filter (FilterComponents | None, optional): The filter of the column. Defaults to None.
            editable (bool | None, optional): Whether the column is editable. Defaults to None.

        Raises:
            Exception: If the column style is not set.
        """
        # Convert the input to the correct format
        params = {}
        if width is not None:
            params["columnWidth"] = width
        if cell_renderer is not None:
            params["cellRenderer"] = cell_renderer.value
        if filter is not None:
            params["filter"] = filter.value
        if editable is not None:
            params["editable"] = editable
        status, message = await AgenticTable.set_column_metadata(
            user_id=self._user_id,
            company_id=self._company_id,
            tableId=self.table_id,
            columnOrder=column,
            **params,
        )
        if status:
            return
        raise Exception(message)

    async def get_num_rows(self) -> int:
        """
        Gets the number of rows in the Agentic Table.

        Returns:
            int: The number of rows in the Agentic Table.
        """
        sheet_info = await AgenticTable.get_sheet_data(
            user_id=self._user_id,
            company_id=self._company_id,
            tableId=self.table_id,
            includeRowCount=True,
            includeCells=False,
            includeLogHistory=False,
        )
        return sheet_info["magicTableRowCount"]

    async def get_sheet(
        self,
        start_row: int = 0,
        end_row: int | None = None,
        batch_size: int = 100,
        include_log_history: bool = False,
        include_cell_meta_data: bool = False,
        include_row_metadata: bool = False,
    ) -> MagicTableSheet:
        """
        Gets the sheet data from the Agentic Table paginated by batch_size.

        Args:
            start_row (int): The start row (inclusive).
            end_row (int | None): The end row (not inclusive).
            batch_size (int): The batch size.
            include_log_history (bool): Whether to include the log history.
            include_cell_meta_data (bool): Whether to include the cell metadata (renderer, selection, agreement status).
            include_row_metadata (bool): Whether to include the row metadata (key value pairs).
        Returns:
            MagicTableSheet: The sheet data.
        """
        # Find the total number of rows
        sheet_info = await AgenticTable.get_sheet_data(
            user_id=self._user_id,
            company_id=self._company_id,
            tableId=self.table_id,
            includeRowCount=True,
            includeCells=False,
            includeLogHistory=False,
            includeCellMetaData=False,
        )
        total_rows = sheet_info["magicTableRowCount"]
        if end_row is None or end_row > total_rows:
            end_row = total_rows
        if start_row > end_row:
            raise Exception("Start row is greater than end row")
        if start_row < 0 or end_row < 0:
            raise Exception("Start row or end row is negative")

        # Get the cells
        cells = []
        for row in range(start_row, end_row, batch_size):
            end_row_batch = min(row + batch_size, end_row)
            sheet_partial = await AgenticTable.get_sheet_data(
                user_id=self._user_id,
                company_id=self._company_id,
                tableId=self.table_id,
                includeCells=True,
                includeLogHistory=include_log_history,
                includeRowCount=False,
                includeCellMetaData=include_cell_meta_data,  # renderer, selection, agreement status
                startRow=row,
                endRow=end_row_batch - 1,
            )
            if "magicTableCells" in sheet_partial:
                if include_row_metadata:
                    # If include_row_metadata is true, we need to get the row metadata for each cell.
                    row_metadata_map = {}
                    # TODO: @thea-unique This routine is not efficient and would be nice if we had this data passed on in get_sheet_data.
                    for cell in sheet_partial["magicTableCells"]:
                        row_order = cell.get("rowOrder")  # type: ignore[assignment]
                        if row_order is not None and row_order not in row_metadata_map:
                            column_order = cell.get("columnOrder")  # type: ignore[assignment]
                            self.logger.info(
                                f"Getting row metadata for cell {row_order}, {column_order}"
                            )
                            cell_with_row_metadata = await self.get_cell(
                                row_order,
                                column_order,  # type: ignore[arg-type]
                            )
                            if cell_with_row_metadata.row_metadata:
                                print(cell_with_row_metadata.row_metadata)
                                row_metadata_map[cell_with_row_metadata.row_order] = (
                                    cell_with_row_metadata.row_metadata
                                )
                                cell["rowMetadata"] = (  # type: ignore[assignment]
                                    cell_with_row_metadata.row_metadata
                                )
                    # Assign row_metadata to all cells
                    for cell in sheet_partial["magicTableCells"]:
                        row_order = cell.get("rowOrder")  # type: ignore[assignment]
                        if row_order is not None and row_order in row_metadata_map:
                            cell["rowMetadata"] = row_metadata_map[  # type: ignore[assignment]
                                row_order
                            ]

                cells.extend(sheet_partial["magicTableCells"])

        sheet_info["magicTableCells"] = cells
        return MagicTableSheet.model_validate(sheet_info)

    async def get_sheet_metadata(self) -> list[RowMetadataEntry]:
        """
        Gets the sheet metadata from the Agentic Table.

        Returns:
            list[RowMetadataEntry]: The sheet metadata.
        """
        sheet_info = await AgenticTable.get_sheet_data(
            user_id=self._user_id,
            company_id=self._company_id,
            tableId=self.table_id,
            includeSheetMetadata=True,  # type: ignore[arg-type]
        )
        return [
            RowMetadataEntry.model_validate(metadata)
            for metadata in sheet_info["magicTableSheetMetadata"]
        ]

    async def set_cell_metadata(
        self,
        row: int,
        column: int,
        selected: bool | None = None,
        selection_method: SelectionMethod | None = None,
        agreement_status: AgreementStatus | None = None,
    ) -> None:
        """
        Sets the cell metadata for the Agentic Table.
        NOTE: This is not to be confused with the sheet metadata and is associated rather with selection and agreement status, row locking etc.

        Args:
            row (int): The row index.
            column (int): The column index.
            selected (bool | None): Whether the cell is selected.
            selection_method (SelectionMethod | None): The method of selection.
            agreement_status (AgreementStatus | None): The agreement status.
        """
        params = {}
        if selected is not None:
            params["selected"] = selected
        if selection_method is not None:
            params["selectionMethod"] = selection_method
        if agreement_status is not None:
            params["agreementStatus"] = agreement_status
        result = await AgenticTable.set_cell_metadata(
            user_id=self._user_id,
            company_id=self._company_id,
            tableId=self.table_id,
            rowOrder=row,
            columnOrder=column,
            **params,
        )
        if result["status"]:  # type: ignore
            return
        raise Exception(result["message"])  # type: ignore

    async def update_row_verification_status(
        self,
        row_orders: list[int],
        status: RowVerificationStatus,
    ):
        """Update the verification status of multiple rows at once.

        Args:
            row_orders (list[int]): The row indexes to update.
            status (RowVerificationStatus): The verification status to set.
        """
        await AgenticTable.bulk_update_status(
            user_id=self._user_id,
            company_id=self._company_id,
            tableId=self.table_id,
            rowOrders=row_orders,
            status=status,
        )



from typing import Any

from jinja2 import Template

from unique_toolkit._common.utils.jinja.schema import Jinja2PromptParams


def render_template(
    template: str, params: Jinja2PromptParams | dict[str, Any] | None = None, **kwargs
) -> str:
    params = params or {}

    if isinstance(params, Jinja2PromptParams):
        params = params.model_dump(exclude_none=True, mode="json")

    params.update(kwargs)

    return Template(template, lstrip_blocks=True).render(**params)



import re
from enum import StrEnum
from typing import Annotated, Generic, Literal, TypeVar

from pydantic import Field
from pydantic.main import BaseModel

from unique_toolkit._common.pydantic.rjsf_tags import RJSFMetaTag
from unique_toolkit._common.pydantic_helpers import get_configuration_dict
from unique_toolkit.agentic.tools.schemas import BaseToolConfig


class SubAgentSystemReminderType(StrEnum):
    FIXED = "fixed"
    REGEXP = "regexp"
    REFERENCE = "reference"
    NO_REFERENCE = "no_reference"


T = TypeVar("T", bound=SubAgentSystemReminderType)


class SystemReminderConfig(BaseModel, Generic[T]):
    model_config = get_configuration_dict()

    type: T


_SYSTEM_REMINDER_FIELD_DESCRIPTION = """
The reminder to add to the tool response. The reminder can be a Jinja template and can contain the following placeholders:
- {{ display_name }}: The display name of the sub agent.
- {{ tool_name }}: The tool name. 
""".strip()


class NoReferenceSystemReminderConfig(SystemReminderConfig):
    """A system reminder that is only added if the sub agent response does not contain any references."""

    type: Literal[SubAgentSystemReminderType.NO_REFERENCE] = (
        SubAgentSystemReminderType.NO_REFERENCE
    )
    reminder: str = Field(
        default="Do NOT create any references from this sub agent in your response! The sub agent response does not contain any references.",
        description=_SYSTEM_REMINDER_FIELD_DESCRIPTION,
    )


class ReferenceSystemReminderConfig(SystemReminderConfig):
    type: Literal[SubAgentSystemReminderType.REFERENCE] = (
        SubAgentSystemReminderType.REFERENCE
    )
    reminder: str = Field(
        default="Rememeber to properly reference EACH fact from sub agent {{ display_name }}'s response with the correct format INLINE. You MUST COPY THE REFERENCE AS PRESENT IN THE SUBAGENT RESPONSE.",
        description=_SYSTEM_REMINDER_FIELD_DESCRIPTION,
    )


class FixedSystemReminderConfig(SystemReminderConfig):
    type: Literal[SubAgentSystemReminderType.FIXED] = SubAgentSystemReminderType.FIXED
    reminder: str = Field(
        description=_SYSTEM_REMINDER_FIELD_DESCRIPTION,
    )


_REGEXP_DETECTED_REMINDER_FIELD_DESCRIPTION = """
The reminder to add to the tool response. The reminder can be a Jinja template and can contain the following placeholders:
- {{ display_name }}: The display name of the sub agent.
- {{ tool_name }}: The tool name. 
- {{ text_matches }}: Will be replaced with the portions of the text that triggered the reminder.
""".strip()


class RegExpDetectedSystemReminderConfig(SystemReminderConfig):
    """A system reminder that is only added if the sub agent response matches a regular expression."""

    type: Literal[SubAgentSystemReminderType.REGEXP] = SubAgentSystemReminderType.REGEXP

    regexp: re.Pattern[str] = Field(
        description="The regular expression to use to detect whether the system reminder should be added.",
    )
    reminder: str = Field(
        description=_REGEXP_DETECTED_REMINDER_FIELD_DESCRIPTION,
    )


SystemReminderConfigType = (
    FixedSystemReminderConfig
    | RegExpDetectedSystemReminderConfig
    | ReferenceSystemReminderConfig
    | NoReferenceSystemReminderConfig
)

DEFAULT_PARAM_DESCRIPTION_SUB_AGENT_USER_MESSAGE = """
This is the message that will be sent to the sub-agent.
""".strip()


class SubAgentToolConfig(BaseToolConfig):
    model_config = get_configuration_dict()

    assistant_id: str = Field(
        default="",
        description="The unique identifier of the assistant to use for the sub-agent.",
    )
    chat_id: str | None = Field(
        default=None,
        description="The chat ID to use for the sub-agent conversation. If None, a new chat will be created.",
    )
    reuse_chat: bool = Field(
        default=True,
        description="Whether to reuse the existing chat or create a new one for each sub-agent call.",
    )
    use_sub_agent_references: bool = Field(
        default=True,
        description="Whether this sub agent's references should be used in the main agent's response.",
    )
    forced_tools: list[str] | None = Field(
        default=None,
        description="The list of tool names that will be forced to be called for this sub-agent.",
    )

    tool_description_for_system_prompt: Annotated[
        str, RJSFMetaTag.StringWidget.textarea(rows=5)
    ] = Field(
        default="",
        description="Description of the tool that will be included in the system prompt.",
    )
    tool_description: Annotated[str, RJSFMetaTag.StringWidget.textarea(rows=5)] = Field(
        default="",
        description="Description of the tool that will be included in the tools sent to the model.",
    )
    param_description_sub_agent_user_message: Annotated[
        str, RJSFMetaTag.StringWidget.textarea(rows=5)
    ] = Field(
        default=DEFAULT_PARAM_DESCRIPTION_SUB_AGENT_USER_MESSAGE,
        description="Description of the user message parameter that will be sent to the model.",
    )
    tool_format_information_for_system_prompt: Annotated[
        str, RJSFMetaTag.StringWidget.textarea(rows=5)
    ] = Field(
        default="",
        description="Format information that will be included in the system prompt to guide response formatting.",
    )
    tool_description_for_user_prompt: Annotated[
        str, RJSFMetaTag.StringWidget.textarea(rows=5)
    ] = Field(
        default="",
        description="Description of the tool that will be included in the user prompt.",
    )
    tool_format_information_for_user_prompt: Annotated[
        str, RJSFMetaTag.StringWidget.textarea(rows=5)
    ] = Field(
        default="",
        description="Format information that will be included in the user prompt to guide response formatting.",
    )

    poll_interval: float = Field(
        default=1.0,
        description="Time interval in seconds between polling attempts when waiting for sub-agent response.",
    )
    max_wait: float = Field(
        default=120.0,
        description="Maximum time in seconds to wait for the sub-agent response before timing out.",
    )
    stop_condition: Literal["stoppedStreamingAt", "completedAt"] = Field(
        default="completedAt",
        description="The condition that will be used to stop the polling for the sub-agent response.",
    )

    tool_input_json_schema: (
        Annotated[str, RJSFMetaTag.StringWidget.textarea(rows=5)] | None
    ) = Field(
        default=None,
        description="A custom JSON schema to send to the llm as the tool input schema.",
    )
    returns_content_chunks: bool = Field(
        default=False,
        description="If set, the sub-agent response will be interpreted as a list of content chunks.",
    )

    system_reminders_config: list[
        Annotated[
            SystemReminderConfigType,
            Field(discriminator="type"),
        ]
    ] = Field(
        default=[],
        description="Configuration for the system reminders to add to the tool response.",
    )



"""Configuration for generation handler."""

from pydantic import BaseModel, Field

from unique_toolkit._common.experimental.write_up_agent.services.generation_handler.prompts.config import (
    GenerationHandlerPromptsConfig,
)
from unique_toolkit._common.pydantic_helpers import get_configuration_dict
from unique_toolkit._common.validators import LMI, get_LMI_default_field
from unique_toolkit.language_model.default_language_model import DEFAULT_GPT_4o


class GenerationHandlerConfig(BaseModel):
    """Configuration for generation handler.

    This configuration controls how groups are batched, how prompts are built,
    and how the LLM is called for generating summaries.
    """

    model_config = get_configuration_dict()

    language_model: LMI = get_LMI_default_field(
        DEFAULT_GPT_4o,
        description="The language model to use for generating summaries.",
    )

    common_instruction: str = Field(
        default="You are a technical writer. Summarize the provided content concisely and clearly.",
        description="Common instruction applied to all groups",
    )

    # TODO [UN-16142]: Add default instructions for each group
    group_specific_instructions: dict[str, str] = Field(
        default_factory=dict,
        description=(
            "Custom instructions per group. "
            "Keys should be formatted as 'column:value' (e.g., 'section:Introduction')"
        ),
    )

    max_tokens_per_batch: int = Field(
        default=4000,
        ge=100,
        description="Maximum tokens per batch for LLM input (affects batching strategy)",
    )

    max_rows_per_batch: int = Field(
        default=20,
        ge=1,
        description="Maximum rows per batch (secondary limit to tokens)",
    )

    prompts_config: GenerationHandlerPromptsConfig = Field(
        default_factory=GenerationHandlerPromptsConfig,
        description="Configuration for the prompts.",
    )



from unique_toolkit.protocols.support import ResponsesSupportCompleteWithReferences
from unique_toolkit.services.chat_service import ChatService


class ResponsesStreamingHandler(ResponsesSupportCompleteWithReferences):
    def __init__(self, chat_service: ChatService):
        self._chat_service = chat_service

    def complete_with_references(self, *args, **kwargs):
        return self._chat_service.complete_responses_with_references(*args, **kwargs)

    async def complete_with_references_async(self, *args, **kwargs):
        return await self._chat_service.complete_responses_with_references_async(
            *args, **kwargs
        )



from typing import Any

from openai.types.chat import (
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartInputAudioParam,
    ChatCompletionContentPartParam,
    ChatCompletionContentPartTextParam,
)
from openai.types.chat.chat_completion_content_part_param import (
    File as ChatCompletionContentPartFileParam,
)
from openai.types.responses import (
    ResponseInputAudioParam,
    ResponseInputFileParam,
    ResponseInputImageParam,
    ResponseInputMessageContentListParam,
    ResponseInputTextParam,
)
from pydantic import (
    TypeAdapter,
)
from pydantic.alias_generators import to_camel


def _convert_text_part_to_responses_api(
    part: ChatCompletionContentPartTextParam,
) -> ResponseInputTextParam:
    return ResponseInputTextParam(type="input_text", text=part["text"])


def _convert_image_part_to_responses_api(
    part: ChatCompletionContentPartImageParam,
) -> ResponseInputImageParam:
    detail = "auto"  # Responses API requires detail to be sent
    if "detail" in part["image_url"]:
        detail = part["image_url"]["detail"]

    return ResponseInputImageParam(
        type="input_image",
        image_url=part["image_url"]["url"],
        detail=detail,
    )


def _convert_file_part_to_responses_api(
    part: ChatCompletionContentPartFileParam,
) -> ResponseInputFileParam:
    param = ResponseInputFileParam(type="input_file")

    if "file_id" in part["file"]:
        param["file_id"] = part["file"]["file_id"]

    if "file_data" in part["file"]:
        param["file_data"] = part["file"]["file_data"]

    if "filename" in part["file"]:
        param["filename"] = part["file"]["filename"]

    return param


def _convert_audio_part_to_responses_api(
    part: ChatCompletionContentPartInputAudioParam,
) -> ResponseInputAudioParam:
    return ResponseInputAudioParam(type="input_audio", input_audio=part["input_audio"])


def convert_user_message_content_to_responses_api(
    content: str | list[dict[str, Any]],
) -> str | ResponseInputMessageContentListParam:
    """Convert user message content from Chat Completions format to Responses API format."""

    if isinstance(content, str):
        return content

    # Parse the content of the dicts in order to allow for camelCase arguments
    content_parsed = TypeAdapter(
        list[ChatCompletionContentPartParam], config={"alias_generator": to_camel}
    ).validate_python(content, by_alias=True, by_name=True)

    responses_content: ResponseInputMessageContentListParam = []

    for part in content_parsed:
        match part["type"]:
            case "text":
                responses_content.append(_convert_text_part_to_responses_api(part))
            case "image_url":
                responses_content.append(_convert_image_part_to_responses_api(part))
            case "file":
                responses_content.append(_convert_file_part_to_responses_api(part))
            case "input_audio":
                responses_content.append(_convert_audio_part_to_responses_api(part))
    return responses_content



"""
Test suite for DebugInfoManager class.

This test suite validates the DebugInfoManager's ability to:
1. Initialize with empty debug info
2. Extract tool debug info from ToolCallResponse objects
3. Handle loop iteration indices
4. Add arbitrary key-value pairs to debug info
5. Retrieve the complete debug info dictionary
"""

from unique_toolkit.agentic.debug_info_manager.debug_info_manager import (
    DebugInfoManager,
)
from unique_toolkit.agentic.tools.schemas import ToolCallResponse


class TestDebugInfoManager:
    """Test suite for DebugInfoManager functionality."""

    def test_init__initializes_empty_debug_info__on_creation(self):
        """Test that DebugInfoManager initializes with empty tools list."""
        manager = DebugInfoManager()

        assert manager.debug_info == {"tools": []}
        assert manager.get() == {"tools": []}

    def test_extract_tool_debug_info__adds_single_tool__with_valid_response(self):
        """Test extracting debug info from a single ToolCallResponse."""
        manager = DebugInfoManager()
        tool_call_response = ToolCallResponse(
            id="tool_1",
            name="TestTool",
            debug_info={"execution_time": "100ms", "status": "success"},
        )

        manager.extract_tool_debug_info([tool_call_response])

        debug_info = manager.get()
        assert len(debug_info["tools"]) == 1
        assert debug_info["tools"][0]["name"] == "TestTool"
        assert debug_info["tools"][0]["info"]["execution_time"] == "100ms"
        assert debug_info["tools"][0]["info"]["status"] == "success"

    def test_extract_tool_debug_info__adds_multiple_tools__with_multiple_responses(
        self,
    ):
        """Test extracting debug info from multiple ToolCallResponse objects."""
        manager = DebugInfoManager()
        tool_call_responses = [
            ToolCallResponse(
                id="tool_1",
                name="SearchTool",
                debug_info={"query": "test query", "results": 5},
            ),
            ToolCallResponse(
                id="tool_2",
                name="CalculatorTool",
                debug_info={"operation": "add", "result": 42},
            ),
            ToolCallResponse(
                id="tool_3",
                name="WeatherTool",
                debug_info={"location": "New York", "temperature": "72F"},
            ),
        ]

        manager.extract_tool_debug_info(tool_call_responses)

        debug_info = manager.get()
        assert len(debug_info["tools"]) == 3
        assert debug_info["tools"][0]["name"] == "SearchTool"
        assert debug_info["tools"][1]["name"] == "CalculatorTool"
        assert debug_info["tools"][2]["name"] == "WeatherTool"

    def test_extract_tool_debug_info__preserves_order__with_sequential_calls(self):
        """Test that multiple calls to extract_tool_debug_info preserve order."""
        manager = DebugInfoManager()

        # First call
        manager.extract_tool_debug_info(
            [ToolCallResponse(id="tool_1", name="Tool1", debug_info={"step": 1})]
        )

        # Second call
        manager.extract_tool_debug_info(
            [ToolCallResponse(id="tool_2", name="Tool2", debug_info={"step": 2})]
        )

        # Third call
        manager.extract_tool_debug_info(
            [ToolCallResponse(id="tool_3", name="Tool3", debug_info={"step": 3})]
        )

        debug_info = manager.get()
        assert len(debug_info["tools"]) == 3
        assert debug_info["tools"][0]["info"]["step"] == 1
        assert debug_info["tools"][1]["info"]["step"] == 2
        assert debug_info["tools"][2]["info"]["step"] == 3

    def test_extract_tool_debug_info__adds_loop_iteration__when_index_provided(self):
        """Test that loop_iteration_index is added to debug info when provided."""
        manager = DebugInfoManager()
        tool_call_response = ToolCallResponse(
            id="tool_1", name="IterativeTool", debug_info={"status": "processing"}
        )

        manager.extract_tool_debug_info([tool_call_response], loop_iteration_index=3)

        debug_info = manager.get()
        assert debug_info["tools"][0]["info"]["loop_iteration"] == 3
        assert debug_info["tools"][0]["info"]["status"] == "processing"

    def test_extract_tool_debug_info__omits_loop_iteration__when_index_is_none(self):
        """Test that loop_iteration is not added when index is None."""
        manager = DebugInfoManager()
        tool_call_response = ToolCallResponse(
            id="tool_1", name="SingleRunTool", debug_info={"status": "complete"}
        )

        manager.extract_tool_debug_info([tool_call_response], loop_iteration_index=None)

        debug_info = manager.get()
        assert "loop_iteration" not in debug_info["tools"][0]["info"]
        assert debug_info["tools"][0]["info"]["status"] == "complete"

    def test_extract_tool_debug_info__handles_empty_debug_info__gracefully(self):
        """Test extracting from ToolCallResponse with empty debug_info dict."""
        manager = DebugInfoManager()
        tool_call_response = ToolCallResponse(
            id="tool_1", name="MinimalTool", debug_info={}
        )

        manager.extract_tool_debug_info([tool_call_response])

        debug_info = manager.get()
        assert len(debug_info["tools"]) == 1
        assert debug_info["tools"][0]["name"] == "MinimalTool"
        assert debug_info["tools"][0]["info"] == {}

    def test_extract_tool_debug_info__includes_mcp_server__when_present_in_debug_info(
        self,
    ):
        """Test that mcp_server is promoted from debug_info to the tool level."""
        manager = DebugInfoManager()
        tool_call_response = ToolCallResponse(
            id="tool_1",
            name="MCPSearchTool",
            debug_info={
                "mcp_tool": "MCPSearchTool",
                "mcp_server": "my-mcp-server",
                "arguments": {"query": "test"},
            },
        )

        manager.extract_tool_debug_info([tool_call_response])

        debug_info = manager.get()
        assert len(debug_info["tools"]) == 1
        assert debug_info["tools"][0]["name"] == "MCPSearchTool"
        assert debug_info["tools"][0]["mcp_server"] == "my-mcp-server"

    def test_extract_tool_debug_info__omits_mcp_server__when_not_present(self):
        """Test that mcp_server is not included for non-MCP tools."""
        manager = DebugInfoManager()
        tool_call_response = ToolCallResponse(
            id="tool_1",
            name="InternalSearch",
            debug_info={"query": "test"},
        )

        manager.extract_tool_debug_info([tool_call_response])

        debug_info = manager.get()
        assert len(debug_info["tools"]) == 1
        assert "mcp_server" not in debug_info["tools"][0]

    def test_extract_tool_debug_info__handles_empty_list__without_error(self):
        """Test that passing an empty list doesn't cause errors."""
        manager = DebugInfoManager()

        manager.extract_tool_debug_info([])

        debug_info = manager.get()
        assert debug_info["tools"] == []

    def test_add__adds_new_key_value_pair__to_debug_info(self):
        """Test adding a new key-value pair to debug_info."""
        manager = DebugInfoManager()

        manager.add("execution_summary", {"total_time": "500ms", "total_calls": 5})

        debug_info = manager.get()
        assert "execution_summary" in debug_info
        assert debug_info["execution_summary"]["total_time"] == "500ms"
        assert debug_info["execution_summary"]["total_calls"] == 5

    def test_add__preserves_tools_list__when_adding_new_keys(self):
        """Test that add() preserves the tools list."""
        manager = DebugInfoManager()
        manager.extract_tool_debug_info(
            [
                ToolCallResponse(
                    id="tool_1", name="TestTool", debug_info={"test": "data"}
                )
            ]
        )

        manager.add("metadata", {"version": "1.0"})

        debug_info = manager.get()
        assert len(debug_info["tools"]) == 1
        assert debug_info["tools"][0]["name"] == "TestTool"
        assert debug_info["metadata"]["version"] == "1.0"

    def test_add__overwrites_existing_key__when_key_exists(self):
        """Test that add() overwrites an existing key."""
        manager = DebugInfoManager()
        manager.add("status", "in_progress")
        manager.add("status", "completed")

        debug_info = manager.get()
        assert debug_info["status"] == "completed"

    def test_add__adds_multiple_keys__with_sequential_calls(self):
        """Test adding multiple key-value pairs with sequential calls."""
        manager = DebugInfoManager()

        manager.add("key1", "value1")
        manager.add("key2", {"nested": "value2"})
        manager.add("key3", [1, 2, 3])

        debug_info = manager.get()
        assert debug_info["key1"] == "value1"
        assert debug_info["key2"]["nested"] == "value2"
        assert debug_info["key3"] == [1, 2, 3]

    def test_get__returns_complete_debug_info__with_mixed_data(self):
        """Test get() returns complete debug info with tools and custom keys."""
        manager = DebugInfoManager()

        # Add tool debug info
        manager.extract_tool_debug_info(
            [ToolCallResponse(id="tool_1", name="Tool1", debug_info={"data": "test"})],
            loop_iteration_index=0,
        )

        # Add custom keys
        manager.add("start_time", "2025-10-16T10:00:00")
        manager.add("end_time", "2025-10-16T10:01:00")

        debug_info = manager.get()

        assert "tools" in debug_info
        assert "start_time" in debug_info
        assert "end_time" in debug_info
        assert len(debug_info["tools"]) == 1
        assert debug_info["start_time"] == "2025-10-16T10:00:00"

    def test_integration__complete_workflow__with_all_operations(self):
        """Integration test: complete workflow using all DebugInfoManager methods."""
        manager = DebugInfoManager()

        # Initial state
        assert manager.get() == {"tools": []}

        # Add some metadata
        manager.add("session_id", "abc-123")
        manager.add("user_id", "user-456")

        # First tool call (loop iteration 0)
        manager.extract_tool_debug_info(
            [
                ToolCallResponse(
                    id="tool_1",
                    name="SearchTool",
                    debug_info={"query": "AI research", "hits": 100},
                )
            ],
            loop_iteration_index=0,
        )

        # Second tool call (loop iteration 1)
        manager.extract_tool_debug_info(
            [
                ToolCallResponse(
                    id="tool_2",
                    name="AnalysisTool",
                    debug_info={"processed": 50, "relevant": 10},
                ),
                ToolCallResponse(
                    id="tool_3",
                    name="SummaryTool",
                    debug_info={"paragraphs": 3, "words": 250},
                ),
            ],
            loop_iteration_index=1,
        )

        # Add final summary
        manager.add("summary", {"total_tools": 3, "total_iterations": 2})

        # Verify complete debug info
        debug_info = manager.get()

        assert debug_info["session_id"] == "abc-123"
        assert debug_info["user_id"] == "user-456"
        assert len(debug_info["tools"]) == 3
        assert debug_info["tools"][0]["name"] == "SearchTool"
        assert debug_info["tools"][0]["info"]["loop_iteration"] == 0
        assert debug_info["tools"][1]["name"] == "AnalysisTool"
        assert debug_info["tools"][1]["info"]["loop_iteration"] == 1
        assert debug_info["tools"][2]["name"] == "SummaryTool"
        assert debug_info["tools"][2]["info"]["loop_iteration"] == 1
        assert debug_info["summary"]["total_tools"] == 3
